"""
Module contains base settings classes
"""
import asyncio
import configparser
import os
import pprint
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Any, Dict, List, NamedTuple, Optional, Set, Union

import dpath.util
import fastjsonschema

from .all_settings import SETTINGS
from .functions import (
    getConfigFromFile,
    getConfiguratorAddressFromFile,
    getSettingsFromConfigurator,
    isConfiguratorEnableInConfigFileOrEnv,
    parseConfiguratorAddress,
    pullConfigFromConfigurator,
)
from .setting_getter import ConfigSource, SettingGetter, getDefaultFromSetting, getSchemaFromSetting

OTHER_SECTION = "OTHER"


class SettingValidationError(Exception):
    """Settings validation exception"""


class ConfiguratorAddress(NamedTuple):
    """
    Structure for storing configurator address which is source of settings
    """

    origin: Union[str, None]  #: origin (or none if settings from file)
    apiVersion: Union[int, None]  #: configurator api version (or none if settings from file)


class _BaseSettingsClass(ABC):
    """
    Base service settings class
    Attributes:
        _source(ConfigSource): source of settings
        _configFile (Option[str]): config file path
        _configuratorAddress (Optional[str]): configurator address
        _settings (Optional[dict]): loaded settings (from file, configurator or default)
        _tags (Optional[Dict[str,str]]): setting tags

    """

    @property
    @abstractmethod
    def settingsStore(self) -> dict:
        """service settings definition store"""

    @abstractmethod
    def _addNonReceivedSettings(self, inputSettingDict: dict) -> None:
        """
        Add settings which non received from a configurator or a configuration file

        Args:
            inputSettingDict - input json with settings
        """

    @abstractmethod
    async def _getSettingsOfInterestFromConfigurator(self, configuratorOrigin: str, configuratorApi: int) -> dict:
        """
        Get config from configurator
        Args:
            configuratorOrigin: configurator origin
            configuratorApi: configurator api version
        Returns:
            dict with config from configurator
        """

    def __init__(self) -> None:
        """
        Initialize. Load default settings
        """

        self._source: ConfigSource = ConfigSource.default
        self._configFile: Union[str, None] = None
        self._configuratorAddress: Union[str, None] = None
        self._settings: Union[Dict, None] = None
        self._loadedConfig: Union[dict, configparser.ConfigParser, None] = None
        self._tags: Union[Dict[str, str], None] = None
        # load default
        self.loadDefault()

    def load(
        self,
        *,
        configFile: Optional[str] = None,
        configuratorAddress: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        asyncLoad=False,
    ):
        """
        Load settings from configurator, configuration file or defaults
        Args:
            configFile: configuration file path
            configuratorAddress: luna configurator address
            tags: tags for load settings from configurator
            asyncLoad: use asynchronous or synchronous load
        """
        loadCoro = self._load(configFile=configFile, configuratorAddress=configuratorAddress, tags=tags)
        if not asyncLoad:
            return asyncio.get_event_loop().run_until_complete(loadCoro)
        return loadCoro

    async def _load(
        self,
        *,
        configFile: Optional[str] = None,
        configuratorAddress: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        """
        Load settings from configurator, configuration file or defaults
        Args:
            configFile: configuration file pat
            configuratorAddress: luna configurator address
            tags: tags for load settings from configurator
        """
        self._configFile = configFile
        self._configuratorAddress = configuratorAddress
        self._tags = tags

        if configuratorAddress is not None:
            self._source = ConfigSource.configurator
            configuratorOrigin, configuratorApi = parseConfiguratorAddress(configuratorAddress)
            self._loadedConfig = await self._getSettingsOfInterestFromConfigurator(configuratorOrigin, configuratorApi)
        elif configFile is not None:
            self._loadedConfig = getConfigFromFile(configFile)
            self._source = (
                ConfigSource.configurator
                if isConfiguratorEnableInConfigFileOrEnv(configFile)
                else ConfigSource.configFile
            )
            if self.isConfigFromConfigurator():
                if self._loadedConfig.has_section("LUNA_CONFIGURATOR"):
                    configuratorSection = dict(self._loadedConfig["LUNA_CONFIGURATOR"])
                else:
                    configuratorSection = {}
                configuratorOrigin, configuratorApi = getConfiguratorAddressFromFile(configuratorSection)
                configuratorOrigin = (
                    os.getenv("VL_SETTINGS.LUNA_CONFIGURATOR.LUNA_CONFIGURATOR_ORIGIN") or configuratorOrigin
                )
                configuratorApi = (
                    os.getenv("VL_SETTINGS.LUNA_CONFIGURATOR.LUNA_CONFIGURATOR_API") or configuratorApi  # type: ignore
                )
                self._configuratorAddress = f"{configuratorOrigin}/{configuratorApi}"
                self._loadedConfig = await self._getSettingsOfInterestFromConfigurator(
                    configuratorOrigin, configuratorApi
                )
        else:
            # load default
            self._loadedConfig = configparser.ConfigParser()
            self._source = ConfigSource.default

        if self.isConfigFromConfigurator():
            jsonConfig = self._createJsonConfigFromConfigurator()
        else:
            jsonConfig = self._createJsonConfigFromConfigFile()

        self._storeSettings(jsonConfig)
        return self

    def loadDefault(self):
        """
        Load default config
        """
        self._loadedConfig = configparser.ConfigParser()
        self._source = ConfigSource.default
        jsonConfig = self._createJsonConfigFromConfigFile()
        self._storeSettings(jsonConfig)

    def _storeSettings(self, jsonConfig: dict):
        """
        Store settings

        Args:
            jsonConfig: config in json format
        """
        self._checkValidSections(jsonConfig)
        self._generateServiceSettings(jsonConfig)
        self._settings = jsonConfig

    def _createJsonConfigFromConfigurator(self) -> Dict:
        """
        Prepares data from the configurator for validation according to the scheme

        Returns:
            prepared config JSON format
        """
        if self._loadedConfig is None or isinstance(self._loadedConfig, configparser.ConfigParser):
            raise RuntimeError("Config was not pulled from configurator")
        preparedConfig = {}
        for sectionName, section in self._loadedConfig.items():
            value = section.get("value")
            preparedConfig[sectionName] = value
        return preparedConfig

    def isConfigFromConfigurator(self) -> bool:
        """
        Get config from luna-configurator or (config file + defaults)

        Returns:
            True if source of settings is config file otherwise False
        """
        return self._source == ConfigSource.configurator

    def _createJsonConfigFromConfigFile(self) -> dict:
        """
        Creates JSON (python-dict) config from settings .conf file
        and adds standard values

        Returns:
            prepared and updated from all_settings.py config
        """
        if not isinstance(self._loadedConfig, configparser.ConfigParser):
            raise RuntimeError("Config was not load from defaults or configuration file")
        settingDict = {}

        for section in self._loadedConfig.sections():
            if section == "OTHER":
                for settingName in self._loadedConfig[section]:
                    settingName = settingName.upper()
                    settingValue = self._loadedConfig[section][settingName]
                    settingDict[settingName] = settingValue
            else:
                # it is required to check digits due to diff from
                # https://github.com/dpath-maintainers/dpath-python/compare/v2.1.0...v2.1.1
                path = [int(value) if value.isdigit() else value for value in section.split(".")]
                dpath.util.new(obj=settingDict, path=path, value=dict(self._loadedConfig[section]))

        return settingDict

    def _checkValidSections(self, settingDict: dict):
        """
        Check if settings sections exist

        Args:
            settingDict - dict with settings
        Raises:
            KeyError: Not find section in settings
        """
        for section in settingDict:
            if section not in self.settingsStore:
                raise KeyError(f"'{section}' is unknown setting")

    def _convertSettingsToSchemaCaseAndFillFromDefault(self, inputSettingDict: dict):
        """
        Convert path in json to schema case and fill settings from default

        Args:
            inputSettingDict - input json with settings
        """
        for settingName, settingValue in inputSettingDict.items():
            schema = getSchemaFromSetting(self.settingsStore[settingName])
            default = getDefaultFromSetting(self.settingsStore[settingName])
            try:
                inputSettingDict[settingName] = SettingGetter(
                    settingName, settingValue, schema, default, self._source
                ).getSetting()
            except fastjsonschema.exceptions.JsonSchemaValueException as exp:
                msg = f"Validation setting error, setting: {settingName}, reason: {exp.message}"
                raise SettingValidationError(msg) from exp

    def _generateServiceSettings(self, inputSettingDict: dict):
        """
        Generate service settings from input json

        Args:
            inputSettingDict - input json with settings
        """
        self._convertSettingsToSchemaCaseAndFillFromDefault(inputSettingDict)
        self._addNonReceivedSettings(inputSettingDict)

    @staticmethod
    def _replacePassword(key: str, value: Any) -> str:
        """
        Replace password with `*` if field contains password
        Args:
            key: field key
            value: field value
        Returns:
            `*` instead of password if password field otherwise field value
        """
        return "*" * 8 if any(subStr in key.lower() for subStr in ("pass", "secret", "token")) else value

    def getConfigAsFlattenDict(self) -> dict:
        """
        Get config as flatten dict with hidden passwords | prepared as config for json response
        """
        Setting = namedtuple("Setting", ("path", "name", "value"))
        configuratorPath = ["LUNA_CONFIGURATOR"]
        settings = []

        def collectSettingsRecursively(data: Any, path: list) -> None:
            if path == configuratorPath and self.isConfigFromConfigurator():
                address, apiVersion = parseConfiguratorAddress(self._configuratorAddress)  # type: ignore
                settings.append(Setting(configuratorPath, "use_configurator", True))
                settings.append(Setting(configuratorPath, "luna_configurator_origin", address))
                settings.append(Setting(configuratorPath, "luna_configurator_api", apiVersion))
            elif isinstance(data, dict):
                for key, value in data.items():
                    newPath = [key] if path is None else path + [key]
                    collectSettingsRecursively(value, newPath)
            elif isinstance(data, list):
                if not path:
                    raise RuntimeError("Unexpected array is settings root")
                for idx, value in enumerate(data):
                    collectSettingsRecursively(value, path + [str(idx)])
            else:
                if not path:
                    raise RuntimeError("Unexpected settings root")
                settingName, settingPath = path[-1], path[:-1] if len(path) > 1 else ["OTHER"]
                settings.append(Setting(settingPath, settingName, self._replacePassword(settingName, data)))

        collectSettingsRecursively(self._settings, [])

        flattenResult: dict = {}
        for setting in settings:
            sectionName = ".".join(setting.path).upper()
            flattenResult.setdefault(sectionName, {})[setting.name.upper()] = setting.value

        return flattenResult

    def getConfigAsText(self) -> str:
        """
        Get config values with hidden passwords | prepared as config for text/plain response
        """
        result = ""
        for sectionName, sectionData in self.getConfigAsFlattenDict().items():
            dataString = "\n".join([f"{key} = {value}" for key, value in sectionData.items()])
            result += f"[{sectionName.upper()}]\n\n{dataString}\n\n"
        return result

    def printConfigValues(self) -> None:
        """
        Print config values with hidden passwords
        """
        prettyPrinter = pprint.PrettyPrinter(indent=4)
        prettyPrinter.pprint(self.getConfigAsFlattenDict())

    def __getitem__(self, settingName: str) -> Any:
        """
        Get setting by name

        Args:
            settingName: setting name

        Returns:
            setting from _settings
        """
        if self._settings is None:
            raise RuntimeError("Config does not loaded")
        settingName = settingName.upper()
        return self._settings[settingName]

    def __eq__(self, other: Any) -> bool:
        """
        Check settings instances' equality.

        Args:
            other: other Settings instance

        Returns:
            True  if equal
            False if not equal
        Raises:
            TypeError if wrong type
        """
        if isinstance(other, BaseServiceSettingsClass):
            return self._settings == other._settings
        raise TypeError(f"Got wrong type {type(other)}.")


class BaseStorageSettingsClass(_BaseSettingsClass):
    """
    Base storage settings class. It is intended for loading settings which specified as `settingsOfInterest`
    regardless of which services these settings refer to.
    """

    @property
    @abstractmethod
    def settingsOfInterest(self) -> List[str]:
        """Settings names which must be loaded from config-file or configurator"""

    def _addNonReceivedSettings(self, inputSettingDict: dict) -> None:
        """Description see :func:`~_BaseSettingsClass._addNonReceivedSettings`."""
        for settingName, settingValue in self.settingsStore.items():
            if settingName in self.settingsOfInterest and settingName not in inputSettingDict:
                inputSettingDict[settingName] = settingValue["default_value"]

    async def _getSettingsOfInterestFromConfigurator(self, configuratorOrigin: str, configuratorApi: int) -> dict:
        """Description see :func:`~_BaseSettingsClass._getSettingsOfInterestFromConfigurator`."""
        return await getSettingsFromConfigurator(  # type: ignore
            self.settingsOfInterest,
            configuratorOrigin=configuratorOrigin,
            configuratorApiVersion=configuratorApi,
            tags=self._tags,
            asyncPulling=True,
        )

    def getLoadedSettings(self) -> dict:
        """Get loaded settings"""
        if self._settings is None:
            raise RuntimeError("Config not loaded")
        return self._settings


class BaseServiceSettingsClass(_BaseSettingsClass):
    """
    Base storage settings class. It is intended for loading settings for service specified by `serviceName`.

    Attributes:
        _upperServiceName(str): service name in UPPER CASE
    """

    # service settings definition store
    settingsStore = SETTINGS

    @property
    @abstractmethod
    def serviceName(self) -> str:
        """Class should registers service name which settings it contains"""

    def __init__(self) -> None:
        """Initialize. Load default settings"""
        self._upperServiceName: str = self.serviceName.replace("-", "_").upper()
        super().__init__()

    def _addNonReceivedSettings(self, inputSettingDict: dict) -> None:
        """Description see :func:`~_BaseSettingsClass._addNonReceivedSettings`."""
        for settingName, settingValue in self.settingsStore.items():
            if settingName not in inputSettingDict and self.serviceName in settingValue.services:
                inputSettingDict[settingName] = getDefaultFromSetting(settingValue)

    async def _getSettingsOfInterestFromConfigurator(self, configuratorOrigin: str, configuratorApi: int) -> dict:
        """Description see :func:`~_BaseSettingsClass._getSettingsOfInterestFromConfigurator`."""
        return await pullConfigFromConfigurator(  # type: ignore
            self.serviceName,
            configuratorOrigin=configuratorOrigin,
            configuratorApiVersion=configuratorApi,
            tags=self._tags,
            asyncPulling=True,
        )

    @classmethod
    def getServiceSettingNames(cls) -> Set[str]:
        """
        Get all service settings names
        Returns:
            set setting names
        """
        serviceSettingList = {
            settingName
            for settingName, settingValue in cls.settingsStore.items()
            if cls.serviceName in settingValue.services
        }
        return serviceSettingList
