"""
Module contains functions to use for work with settings
"""
import asyncio
import configparser
import os
from typing import Awaitable, List, Optional, Tuple, Union
from urllib.parse import urlparse

from luna3.configurator.configurator import ConfiguratorApi  # pylint: disable-msg=E0401

from configs_core.all_settings import SETTINGS


def parseConfiguratorAddress(configuratorAddress: str) -> Tuple[str, int]:
    """
    Get configurator address from command line arguments.

    Args:
        configuratorAddress: configurator address

    Returns:
        tuple<origin, apiVersion>
    """
    # < scheme >: // < netloc > / < path >; < params >? < query >  # <fragment>
    scheme, netloc, path, *_ = urlparse(configuratorAddress)

    if not netloc:
        # when only ip provided
        netloc, path = path, ""
    scheme = scheme or "http"
    host, port = netloc.split(":") if ":" in netloc else (netloc, 5070)

    configuratorApiVersion = int(path.split("/")[1]) if path else 1
    configuratorOrigin = f"{scheme}://{host}:{port}"
    return configuratorOrigin, configuratorApiVersion


def getConfiguratorAddressFromFile(configuratorFileSection: dict) -> Tuple[str, int]:
    """
    Get configurator address from config file or defaults if not exists
    Args:
        configuratorFileSection: dict with configurator section from config file

    Returns:
        tuple<origin, apiVersion>
    """
    defaults = SETTINGS["LUNA_CONFIGURATOR"].default
    configuratorOrigin = configuratorFileSection.get("luna_configurator_origin", defaults["luna_configurator_origin"])
    configuratorApi = configuratorFileSection.get("luna_configurator_api", defaults["luna_configurator_api"])
    return configuratorOrigin, configuratorApi


def getConfigFromFile(configPath: str) -> configparser.ConfigParser:
    """
    Get config from file (default if not specified)
    Args:
        configPath: path to config file
    Returns:
        config from file
    """
    configParser = configparser.ConfigParser()
    if not os.path.isfile(configPath):
        raise RuntimeError(f"Configuration file {configPath} not found")
    configParser.read(configPath)
    return configParser


def pullConfigFromConfigurator(
    serviceName: str,
    configuratorOrigin: str,
    configuratorApiVersion: int,
    tags: Optional[dict] = None,
    asyncPulling=False,
) -> Union[dict, Awaitable[dict]]:
    """
    Pull config for required service
    Args:
        serviceName: required service name
        configuratorOrigin: configurator origin
        configuratorApiVersion: configurator api version
        tags: map setting to tag for pulling settings from configurator
        asyncPulling: use asynchronous or synchronous pulling

    Returns:
        json from configurator if asyncPulling=False otherwise coroutine with json
    """
    coroutine = _pullConfigFromConfigurator(
        serviceName=serviceName,
        configuratorOrigin=configuratorOrigin,
        configuratorApiVersion=configuratorApiVersion,
        tags=tags,
    )
    if asyncPulling:
        return coroutine
    return asyncio.get_event_loop().run_until_complete(coroutine)


async def _pullConfigFromConfigurator(
    serviceName: str, configuratorOrigin: str, configuratorApiVersion: int, tags: Optional[dict] = None
) -> dict:
    """
    Pull config for required service
    Args:
        serviceName: required service name
        configuratorOrigin: configurator origin
        configuratorApiVersion: configurator api version
        tags: map setting to tag for pulling settings from configurator

    Returns:
        json from configurator
    """
    confClient = ConfiguratorApi(configuratorOrigin, configuratorApiVersion)
    reply = await confClient.pullConfig(serviceName, tags, raiseError=False, asyncRequest=True)
    if not reply.success:
        raise RuntimeError(f"Error while pulling config from luna-configurator. {reply.json}")
    return reply.json


def pullSettingFromConfigurator(
    settingName: str,
    configuratorOrigin: str,
    configuratorApiVersion: Optional[int] = None,
    tags: Optional[str] = None,
    asyncPulling=False,
) -> Union[dict, Awaitable[dict]]:
    """
    Pull config for required service
    Args:
        settingName: required setting name
        configuratorOrigin: configurator origin
        configuratorApiVersion: configurator api version
        tags: set of tag
        asyncPulling: use asynchronous or synchronous pulling

    Returns:
        first setting which has any tag (if tags is not None)
    """
    coroutine = _pullSettingFromConfigurator(
        settingName=settingName,
        configuratorOrigin=configuratorOrigin,
        configuratorApiVersion=configuratorApiVersion,
        tags=tags,
    )
    if asyncPulling:
        return coroutine
    return asyncio.get_event_loop().run_until_complete(coroutine)


async def _pullSettingFromConfigurator(
    settingName: str, configuratorOrigin: str, configuratorApiVersion: Optional[int] = None, tags: Optional[str] = None
) -> dict:
    """
    Pull config for required service
    Args:
        settingName: required setting name
        configuratorOrigin: configurator origin
        configuratorApiVersion: configurator api version
        tags: set of tag

    Returns:
        first setting which has any tag (if tags is not None)
    """
    if configuratorApiVersion:
        confClient = ConfiguratorApi(configuratorOrigin, api=configuratorApiVersion)
    else:
        confClient = ConfiguratorApi(configuratorOrigin)
    if tags:
        isDefault = None
    else:
        isDefault = True
    reply = await confClient.getSettings(
        settingName=settingName, tags=tags, isDefault=isDefault, raiseError=False, asyncRequest=True
    )
    if not reply.success:
        raise RuntimeError(f"Error while pulling config from luna-configurator. {reply.json}")
    if not reply.json["settings"]:
        if tags:
            raise RuntimeError(f"Settings not found, name {settingName}, tags: {tags}")
        raise RuntimeError(f"Settings not found, name {settingName}")
    return reply.json["settings"][0]


def getSettingsFromConfigurator(
    settings: List[str],
    configuratorOrigin: str,
    configuratorApiVersion: int,
    tags: Optional[dict] = None,
    asyncPulling=False,
) -> Union[dict, Awaitable[dict]]:
    """
    Get settings from configurator
    Args:
        settings: settings list
        configuratorOrigin: configurator origin
        configuratorApiVersion: configurator api version
        tags: map setting to tag for pulling settings from configurator
        asyncPulling: use asynchronous or synchronous pulling

    Returns:
        json from configurator if asyncPulling=False otherwise coroutine with json
    """
    coroutine = _getSettingsFromConfigurator(
        settings=settings,
        configuratorOrigin=configuratorOrigin,
        configuratorApiVersion=configuratorApiVersion,
        tags=tags,
    )
    if asyncPulling:
        return coroutine
    return asyncio.get_event_loop().run_until_complete(coroutine)


async def _getSettingsFromConfigurator(
    settings: List[str], configuratorOrigin: str, configuratorApiVersion: int, tags: Optional[dict] = None
) -> dict:
    """
    Pull config for required service
    Args:
        settings: settings list
        configuratorOrigin: configurator origin
        configuratorApiVersion: configurator api version
        tags: map setting to tag for pulling settings from configurator

    Returns:
        json from configurator with setting names as keys and its' values and ids as values
    """
    confClient = ConfiguratorApi(configuratorOrigin, configuratorApiVersion)
    res = {}
    for settingName in settings:
        settingTag = tags.get(settingName) if tags is not None else None
        reply = await confClient.getSettings(
            settingName=settingName,
            tags=settingTag,
            isDefault=settingTag is None,
            asyncRequest=True,
            raiseError=False,
        )
        if not reply.success:
            raise RuntimeError(f"Error while getting settings from luna-configurator. {reply.json}")
        if reply.json["count"] > 0:
            setting = reply.json["settings"][0]
            res[setting["name"]] = dict(value=setting["value"], id=setting["id"])
    return res


def isConfiguratorEnableInConfigFileOrEnv(configFile: str) -> bool:
    """
    Check whether configurator is enable from configuration file or environment variables
    Args:
        configFile: configFile

    Returns:
        True if enabled otherwise False
    """
    configFromFile = getConfigFromFile(configFile)
    return configFromFile.has_section("LUNA_CONFIGURATOR") and bool(
        int(os.getenv("VL_SETTINGS.LUNA_CONFIGURATOR.USE_CONFIGURATOR", "0"))
        or int(configFromFile["LUNA_CONFIGURATOR"].get("USE_CONFIGURATOR", "0"))
    )
