"""
Module contains DatabaseMigrationHelper with functions to add/remove/update limitations/groups/settings
"""
import sys
from dataclasses import dataclass
from logging import Logger
from typing import Any, Callable, Iterable, List, Optional, Set, Tuple, Union

import jsonschema
import ujson
from sqlalchemy import and_, delete, insert, select, text, update
from sqlalchemy.engine import Connection

from . import models
from .classes import ExternalLimitationToModify, LimitationToModify, OperationMode


def assertSchemas(limitationName: str, currentSchema: dict, expectedSchema: dict, logger: Logger):
    """Assert that schemas are equal"""
    for field in ("type", "properties", "additionalProperties", "required"):
        if (currentField := currentSchema.get(field)) != (expectedField := expectedSchema.get(field)):
            logger.error(f"Schema for limitation '{limitationName}' from DB does not match to expected.")
            logger.debug(f"Database limitation '{field}' in schema: {currentField}")
            logger.debug(f"Actual limitation '{field}' in schema: {expectedField}")
            logger.debug(f"Database limitation schema: {currentSchema}")
            logger.debug(f"Actual limitation schema: {expectedSchema}")
            logger.error("Aborting.")
            sys.exit(1)


@dataclass
class DatabaseMigrationHelper:
    """Migration helper"""

    # db connection | must be initialized for each revision migration | hax for pylint
    connection: Connection
    # operation mode - upgrade or downgrade | must be initialized for each revision migration | hax for pylint
    operationMode: OperationMode
    # logger with template contains revision | must be initialized for each revision migration | hax for pylint
    logger: Logger
    # current services | must be initialized for each revision migration | hax for pylint
    availableServices: Optional[List[str]]

    def getDBCurrentTimestamp(self):
        """Get sql function `now`"""
        if self.connection.dialect.name == "postgresql":
            return text("timezone('UTC', now())")
        return text("SYS_EXTRACT_UTC(SYSTIMESTAMP)")

    def addLimitation(
        self,
        limitationName: str,
        services: List[str],
        validationSchema: dict,
        defaultValue: Any,
        description: str,
        groupNames: List[str],
    ) -> None:
        """Add limitation, its' default setting and group links to db"""
        if self.availableServices and not set(services).intersection(set(self.availableServices)):
            self.logger.debug(f"Skip adding limitation '{limitationName}', not found in specified services")
            return
        schemaReply = self.connection.execute(
            select([models.Limitation.validation_schema]).where(models.Limitation.limitation_name == limitationName)
        ).cursor.fetchone()
        if schemaReply:
            currentSchema = ujson.loads(schemaReply[0])
            assertSchemas(limitationName, currentSchema, validationSchema, self.logger)
            self.logger.debug(f"Skip adding limitation '{limitationName}', already exists")
            return
        limitationQuery = insert(models.Limitation).values(  # type: ignore
            limitation_name=limitationName,
            validation_schema=ujson.dumps(validationSchema),
            default_value=ujson.dumps(defaultValue),
            services=",".join(services),
            description=description,
        )
        self.connection.execute(limitationQuery)

        for groupName in groupNames:
            groupQuery = insert(models.GroupLimitation).values(  # type: ignore
                limitation_name=limitationName, group_name=groupName
            )
            self.connection.execute(groupQuery)

        settingQuery = insert(models.Setting).values(  # type: ignore
            value=ujson.dumps(defaultValue), description=description, name=limitationName, is_default=True
        )
        self.connection.execute(settingQuery)
        self.logger.debug(f"Added limitation: '{limitationName}',\n\tit's groups: {groupNames}")

    def removeLimitation(self, limitationName: str):
        """Remove limitation from db"""
        limitationQuery = delete(models.Limitation).where(  # type: ignore
            models.Limitation.limitation_name == limitationName
        )
        self.connection.execute(limitationQuery)
        self.logger.debug(f"Removed limitation: '{limitationName}'")

    def _getUpgradeDowngradeData(self, oldList: Iterable[str], newList: Iterable[str]) -> Tuple[Set[str], Set[str]]:
        """Calculate difference and get tuple with data to remove and data"""
        newData = set(oldList).difference(newList)
        oldData = set(newList).difference(oldList)
        if self.operationMode == OperationMode.upgrade:
            newData, oldData = oldData, newData
        return newData, oldData

    def getLimitation(self, limitationName: str) -> dict:
        """
        Get limitation by name

        Args:
            limitationName: ,imitation name

        Returns:
            dict in format BaseSettingDefinition.asDict (except field `external`)
        Raises:
            RuntimeError: if limitation not found
        """
        selectSt = select([models.Limitation]).where(  # type: ignore
            models.Limitation.limitation_name == limitationName
        )
        record = self.connection.execute(selectSt).cursor.fetchone()
        if not record:
            raise RuntimeError(f"limitation '{limitationName}' not found")
        limitation = {
            "limitation_name": record[0],
            "validation_schema": ujson.loads(record[1]),
            "services": record[2],
            "default_value": record[3],
            "description": record[4],
        }
        selectSt = select([models.GroupLimitation.group_name]).where(
            models.GroupLimitation.limitation_name == limitationName
        )
        records = self.connection.execute(selectSt).cursor.fetchall()
        limitation["group_names"] = [record[0] for record in records]
        return limitation

    def upgradeLimitationServices(
        self,
        limitationName: str,
        servicesPlus: Optional[Iterable[str]] = None,
        servicesMinus: Optional[Iterable[str]] = None,
    ):
        """
        Upgrade limitation services

        Args:
            limitationName: limitation name
            servicesPlus: a set services to add
            servicesMinus: a set services to remove
        """
        oldServices = self._getLimitationServices(limitationName)
        servicesMinus = servicesMinus or set()
        servicesPlus = servicesPlus or set()
        newServices = oldServices.union(set(servicesPlus)) - set(servicesMinus)

        updateSt = (
            update(models.Limitation)  # type: ignore
            .where(models.Limitation.limitation_name == limitationName)
            .values(services=",".join(newServices))
        )
        self.connection.execute(updateSt)

    def upgradeLimitationGroups(
        self,
        limitationName: str,
        groupsPlus: Optional[Iterable[str]] = None,
        groupsMinus: Optional[Iterable[str]] = None,
    ):
        """
        Upgrade limitation groups.
        Args:
            limitationName: limitation name
            groupsPlus: a set groups to add
            groupsMinus: a set groups to remove
        """
        if groupsMinus:
            deleteSt = delete(models.GroupLimitation).where(  # type: ignore
                and_(
                    models.GroupLimitation.limitation_name == limitationName,
                    models.GroupLimitation.group_name.in_(groupsMinus),
                )
            )
            self.connection.execute(deleteSt)
        if groupsPlus:
            # check that groups are existing, if not - creating
            selectSt = select([models.Group.group_name]).where(models.Group.group_name.in_(groupsPlus))
            rows = self.connection.execute(selectSt)
            existingGroups = set((row[0] for row in rows))
            notExistingGroups = set(groupsPlus) - existingGroups
            for group in notExistingGroups:
                insertSt = insert(models.Group).values(description="", group_name=group)  # type: ignore
                self.connection.execute(insertSt)
            # check that a limitation to group link are existing, if not - create

            selectSt = select([models.GroupLimitation.group_name]).where(
                and_(
                    models.GroupLimitation.group_name.in_(groupsPlus),
                    models.GroupLimitation.limitation_name == limitationName,
                )
            )
            rows = self.connection.execute(selectSt)
            linkedGroups = set((row[0] for row in rows))
            notExistingLinks = set(groupsPlus) - linkedGroups
            for group in notExistingLinks:
                insertSt = insert(models.GroupLimitation).values(  # type: ignore
                    limitation_name=limitationName, group_name=group
                )
                self.connection.execute(insertSt)

    def _getLimitationServices(self, limitationName: str) -> set[str]:
        """Get limitation services"""
        selectSt = select([models.Limitation.services]).where(models.Limitation.limitation_name == limitationName)
        record = self.connection.execute(selectSt).cursor.fetchone()
        if not record:
            raise RuntimeError(f"limitation '{limitationName}' not found")
        services = set(record[0].split(","))
        return services

    def _updateServices(self, limitationName: str, oldServices: List[str], newServices: List[str]):
        """Update limitation services in db"""
        servicesToAdd, servicesToRemove = self._getUpgradeDowngradeData(oldServices, newServices)
        services = self._getLimitationServices(limitationName)
        if not set(oldServices).issubset(services):
            self.logger.info(f"Limitation {limitationName} services was modified, {oldServices} will be added")
            servicesToAdd |= set(oldServices)

        services -= servicesToRemove
        services |= servicesToAdd
        updateSt = (
            update(models.Limitation)  # type: ignore
            .where(models.Limitation.limitation_name == limitationName)
            .values(services=",".join(services))
        )
        self.connection.execute(updateSt)

    def _updateDescription(self, limitationName: str, oldDescription: str, newDescription: str):
        """Update limitation description in db"""
        description = newDescription if self.operationMode == OperationMode.upgrade else oldDescription
        self.connection.execute(
            update(models.Limitation)  # type: ignore
            .where(models.Limitation.limitation_name == limitationName)
            .values(description=description)
        )

    def _updateGroupNames(self, limitationName: str, oldGroups: List[str], newGroups: List[str]):
        """Update limitation groups in db"""
        groupsToAdd, groupsToRemove = self._getUpgradeDowngradeData(oldGroups, newGroups)

        selectSt = select([models.GroupLimitation.group_name]).where(
            models.GroupLimitation.limitation_name == limitationName
        )
        groups = {row[0] for row in self.connection.execute(selectSt).cursor.fetchall()}
        if not set(oldGroups).issubset(groups):
            self.logger.debug(f"Limitation '{limitationName}' groups corrupted, update to {newGroups}")
            groupsToAdd, groupsToRemove = self._getUpgradeDowngradeData(groups, newGroups)
        groupsToAdd -= groups
        groupsToRemove &= groups
        if groupsToRemove:
            deleteSt = delete(models.GroupLimitation).where(  # type: ignore
                and_(
                    models.GroupLimitation.limitation_name == limitationName,
                    models.GroupLimitation.group_name.in_(groupsToRemove),
                )
            )
            self.connection.execute(deleteSt)
        for group in groupsToAdd:
            insertSt = insert(models.GroupLimitation).values(  # type: ignore
                limitation_name=limitationName, group_name=group
            )
            self.connection.execute(insertSt)

    def _updateDefaultValue(self, limitationName: str, oldDefaultValue: Any, newDefaultValue: Any):
        """Update limitation default value"""
        defaultValue = newDefaultValue if self.operationMode == OperationMode.upgrade else oldDefaultValue
        self.connection.execute(
            update(models.Limitation)  # type: ignore
            .where(models.Limitation.limitation_name == limitationName)
            .values(default_value=ujson.dumps(defaultValue))
        )

    def _updateValidationSchema(self, limitationName: str, oldValidationSchema: Any, newValidationSchema: Any):
        """Update limitation validation schema"""
        validationSchema = newValidationSchema if self.operationMode == OperationMode.upgrade else oldValidationSchema
        self.connection.execute(
            update(models.Limitation)  # type: ignore
            .where(models.Limitation.limitation_name == limitationName)
            .values(validation_schema=ujson.dumps(validationSchema))
        )

    def _settingUpgradeFunction(
        self,
        limitationName: str,
        newDefaultValue: object,
        validationSchema: Optional[dict] = None,
        settingUpgradeFunction: Optional[Callable] = None,
    ):
        """
        Update setting values
        Args:
            limitationName: limitation (setting) name
            newDefaultValue: new value for settings
            validationSchema: validation schema
            settingUpgradeFunction: value to upgrade setting with (priority is higher than newDefaultValue)
        """
        if settingUpgradeFunction:
            settingIds = [
                row[0]
                for row in self.connection.execute(
                    select([models.Setting.id]).where(models.Setting.name == limitationName)
                )
            ]
            for settingId in settingIds:
                oldValue = ujson.loads(
                    self.connection.execute(
                        select([models.Setting.value]).where(models.Setting.id == settingId)
                    ).cursor.fetchone()[0]
                )
                newValue = settingUpgradeFunction(oldValue)
                if validationSchema is not None:
                    try:
                        jsonschema.validate(newValue, validationSchema)
                    except jsonschema.ValidationError:
                        self.logger.error(
                            f"Error during upgrade/downgrade setting '{limitationName}'. "
                            "Setting value received after 'settingUpgradeFunction' does not match "
                            f"limitation validation schema. Old setting value: {oldValue}"
                        )
                        raise
                self.connection.execute(
                    update(models.Setting)  # type: ignore
                    .where(models.Setting.id == settingId)
                    .values(
                        value=ujson.dumps(newValue),
                        last_update_time=self.getDBCurrentTimestamp(),  # type: ignore # pylint: disable-msg=no-member
                    )
                )
        else:
            if validationSchema is not None:
                try:
                    jsonschema.validate(newDefaultValue, validationSchema)
                except jsonschema.ValidationError:
                    self.logger.error(
                        f"Error during upgrade/downgrade setting '{limitationName}'. "
                        "Setting value specified as the new default value does not match "
                        "limitation validation schema."
                    )
                    raise
            self.connection.execute(
                update(models.Setting)  # type: ignore
                .where(models.Setting.name == limitationName)
                .values(
                    value=ujson.dumps(newDefaultValue),
                    last_update_time=self.getDBCurrentTimestamp(),  # type: ignore # pylint: disable-msg=no-member
                )
            )

    def _migrateExternalLimitation(self, limitation: ExternalLimitationToModify):
        """Upgrade/downgrade limitation"""
        if self.availableServices:
            if all(
                (
                    not set(limitation.newServices).intersection(set(self.availableServices)),
                    not set(limitation.oldServices).intersection(set(self.availableServices)),
                )
            ):
                self.logger.debug(
                    f"Skip modifying limitation '{limitation.limitationName}', not found in specified services"
                )
                return
        if limitation.oldServices != limitation.newServices:
            servicesPlus, servicesMinus = self._getUpgradeDowngradeData(limitation.oldServices, limitation.newServices)
            self.upgradeLimitationServices(limitation.limitationName, servicesPlus, servicesMinus)
        if limitation.newGroupNames != limitation.oldGroupNames:
            if limitation.newGroupNames is None or limitation.oldGroupNames is None:
                raise RuntimeError(f"{limitation.limitationName}: One of groups does not filled")
            groupsPlus, groupsMinus = self._getUpgradeDowngradeData(limitation.oldGroupNames, limitation.newGroupNames)
            self.upgradeLimitationGroups(limitation.limitationName, groupsPlus, groupsMinus)
        self.logger.debug(f"Limitation '{limitation.limitationName}' has been modified")

    def _migrateRegularLimitation(self, limitation: LimitationToModify):
        """Upgrade/downgrade limitation"""
        if self.availableServices and not set(limitation.newServices).intersection(set(self.availableServices)):
            self.logger.debug(
                f"Skip modifying limitation '{limitation.limitationName}', not found in specified services"
            )
            return
        if limitation.oldServices != limitation.newServices:
            self._updateServices(limitation.limitationName, limitation.oldServices, limitation.newServices)
        if limitation.newDescription != limitation.oldDescription:
            if limitation.newDescription is None or limitation.oldDescription is None:
                raise RuntimeError(f"{limitation.limitationName}: One of descriptions does not filled")
            self._updateDescription(limitation.limitationName, limitation.oldDescription, limitation.newDescription)
        if limitation.newGroupNames != limitation.oldGroupNames:
            if limitation.newGroupNames is None or limitation.oldGroupNames is None:
                raise RuntimeError(f"{limitation.limitationName}: One of groups does not filled")
            self._updateGroupNames(limitation.limitationName, limitation.oldGroupNames, limitation.newGroupNames)
        if limitation.oldValidationSchema != limitation.newValidationSchema:
            self._updateValidationSchema(
                limitation.limitationName,
                limitation.oldValidationSchema,
                limitation.newValidationSchema,
            )
        isNeedToUpdateDefaultValue = limitation.oldDefaultValue != limitation.newDefaultValue
        if isNeedToUpdateDefaultValue:
            self._updateDefaultValue(limitation.limitationName, limitation.oldDefaultValue, limitation.newDefaultValue)
        if (
            isNeedToUpdateDefaultValue or limitation.settingUpgradeFunction
        ) and self.operationMode == OperationMode.upgrade:
            self._settingUpgradeFunction(
                limitation.limitationName,
                limitation.newDefaultValue,
                limitation.newValidationSchema,
                limitation.settingUpgradeFunction,
            )
        if (
            isNeedToUpdateDefaultValue or limitation.settingDowngradeFunction
        ) and self.operationMode == OperationMode.downgrade:
            self._settingUpgradeFunction(
                limitation.limitationName,
                limitation.oldDefaultValue,
                limitation.oldValidationSchema,
                limitation.settingDowngradeFunction,
            )
        self.logger.debug(f"Limitation '{limitation.limitationName}' has been modified")

    def migrateLimitation(self, limitation: Union[LimitationToModify, ExternalLimitationToModify]):
        """Upgrade/downgrade limitation"""
        if isinstance(limitation, LimitationToModify):
            return self._migrateRegularLimitation(limitation)
        return self._migrateExternalLimitation(limitation)

    def addGroup(self, groupName: str, groupDescription: str):
        """Add group"""
        groupReply = self.connection.execute(
            select([models.Group.group_name]).where(models.Group.group_name == groupName)
        ).cursor.fetchone()
        if groupReply:
            self.logger.debug(f"Skip adding group '{groupName}', already exists")
            return
        groupQuery = insert(models.Group).values(group_name=groupName, description=groupDescription)  # type: ignore
        self.connection.execute(groupQuery)
        self.logger.debug(f"Added group: '{groupName}'")

    def removeGroup(self, groupName: str):
        """Remove group"""
        groupQuery = delete(models.Group).where(models.Group.group_name == groupName)  # type: ignore
        self.connection.execute(groupQuery)
        self.logger.debug(f"Removed group: '{groupName}'")

    def renameLimitation(self, oldName: str, newName: str):
        """Rename limitation, it's settings, their tags and link to groups"""
        limitationReply = self.connection.execute(
            select([models.Limitation.limitation_name]).where(models.Limitation.limitation_name.in_((oldName, newName)))
        ).cursor.fetchall()
        existingLimitations = [limitation[0] for limitation in limitationReply]
        if not existingLimitations or len(existingLimitations) > 1:
            self.logger.error(f"Incorrect number of limitations in DB for renaming. Limitations: {existingLimitations}")
            self.logger.error("Aborting.")
            sys.exit(1)
        if newName in existingLimitations:
            self.logger.debug(f"Skip rename limitation: '{newName}' already exists")
            return

        limitationStandingColumns = [
            models.Limitation.validation_schema,
            models.Limitation.services,
            models.Limitation.default_value,
            models.Limitation.description,
        ]
        addLimitationQuery = insert(models.Limitation).from_select(  # type: ignore
            [models.Limitation.limitation_name, *limitationStandingColumns],  # type: ignore
            select([text(f"'{newName}'"), *limitationStandingColumns]).where(  # type: ignore
                models.Limitation.limitation_name == oldName
            ),
        )
        self.connection.execute(addLimitationQuery)

        updateSettingsQuery = (
            update(models.Setting)  # type: ignore
            .where(models.Setting.name == oldName)
            .values(name=newName, last_update_time=self.getDBCurrentTimestamp())  # pylint: disable-msg=no-member
        )
        self.connection.execute(updateSettingsQuery)
        updateTagQuery = update(models.Tag).where(models.Tag.name == oldName).values(name=newName)  # type: ignore
        self.connection.execute(updateTagQuery)
        updateGroupQuery = (
            update(models.GroupLimitation)  # type: ignore
            .where(models.GroupLimitation.limitation_name == oldName)
            .values(limitation_name=newName)
        )
        self.connection.execute(updateGroupQuery)

        removeOldLimitationQuery = delete(models.Limitation).where(  # type: ignore
            models.Limitation.limitation_name == oldName
        )
        self.connection.execute(removeOldLimitationQuery)
        self.logger.debug(f"Limitation has been renamed: from '{oldName}' to '{newName}'")

    def updateGroup(self, groupName: str, groupDescription: str):  # pylint: disable-msg=W0613
        """Update group | Not implemented"""
        raise NotImplementedError
