"""
Module contains models for luna-configurator database
"""
from typing import Type

from sqlalchemy import TIMESTAMP, Boolean, Column, ForeignKey, Integer, Sequence, String
from sqlalchemy.ext.declarative import AbstractConcreteBase, declarative_base

Base: Type[AbstractConcreteBase] = declarative_base()


class ConfigsMigration(Base):  # type: ignore # pylint: disable-msg=too-few-public-methods
    """
    Configs migration
    """

    __tablename__ = "configs_migrations"

    # settings migration revision
    revision = Column(String, comment="settings migration revision", primary_key=True)
    # settings migration component
    component = Column(String(36), comment="component name")


class Limitation(Base):  # type: ignore # pylint: disable-msg=too-few-public-methods
    """
    Limitation table
    """

    __tablename__ = "limitation"
    __table_args__ = {"comment": "Table with limitations."}

    # setting name
    limitation_name = Column(String, comment="limitation name", primary_key=True)

    # Validation schema
    validation_schema = Column(String, comment="validation schema")

    # Services for setting
    services = Column(String, comment="list of services")

    # Default value
    default_value = Column(String, comment="default limitation value")

    # description
    description = Column(String, default="", comment="limitation description")

    def __repr__(self):
        return f"<Limitation (name={self.limitation_name})>"


class Setting(Base):  # type: ignore # pylint: disable-msg=too-few-public-methods
    """
    Setting table
    """

    __tablename__ = "setting"
    __table_args__ = {"comment": "Table of settings."}

    # setting id
    id = Column(Integer, Sequence("setting_id", start=1), comment="setting id", primary_key=True)
    # setting value
    value = Column(String, comment="setting value")
    # description
    description = Column(String(128), comment="settings description")
    # create time
    create_time = Column(TIMESTAMP, comment="setting create time")
    # last update time
    last_update_time = Column(TIMESTAMP, comment="setting last update time")
    # setting name
    name = Column(String, comment="setting name")
    # whether a setting is a default setting
    is_default = Column(Boolean, nullable=False, default=False, comment="whether a setting is a default setting")

    def __repr__(self):
        return f"<Setting(id={self.id})>"


class Tag(Base):  # type: ignore # pylint: disable-msg=too-few-public-methods
    """
    Tag table.
    """

    __tablename__ = "tag"
    __table_args__ = {"comment": "Table of tags."}

    # setting id
    id = Column(Integer, ForeignKey(Setting.id, ondelete="CASCADE"), comment="setting id", primary_key=True)
    # setting name
    name = Column(String, primary_key=True, comment="setting name")
    # string with tag
    tag = Column(String, primary_key=True, comment="setting tag")

    def __repr__(self):
        return f"<Tag(id={self.id})>"


class Group(Base):  # type: ignore # pylint: disable-msg=too-few-public-methods
    """
    Group table
    """

    __tablename__ = "group"
    __table_args__ = {"comment": "Table of groups"}

    group_name = Column(String, primary_key=True, comment="Group name")

    description = Column(String, comment="Group description")


class GroupLimitation(Base):  # type: ignore # pylint: disable-msg=too-few-public-methods
    """
    Limitation group link table
    """

    __tablename__ = "group_limitation"
    __table_args__ = {"comment": "Table of limitation groups links"}

    group_name = Column(
        String, ForeignKey(Group.group_name, ondelete="CASCADE"), primary_key=True, comment="Group name"
    )

    limitation_name = Column(
        String,
        ForeignKey(Limitation.limitation_name, ondelete="CASCADE"),
        primary_key=True,
        comment="limitation name",
    )
