"""Module realize a setting load functionality"""
import configparser
import os
from collections import defaultdict
from copy import deepcopy
from dataclasses import _MISSING_TYPE, MISSING, dataclass, field
from enum import Enum
from functools import reduce
from operator import getitem
from pathlib import Path
from typing import Generic, Optional, Type, TypeVar, Union

import dpath.util
import fastjsonschema
from pydantic.tools import _get_parsing_type  # pylint: disable-msg=no-name-in-module

from .base_setting_definition import BaseSettingDefinition
from .functions import (
    getConfigFromFile,
    getConfiguratorAddressFromFile,
    isConfiguratorEnableInConfigFileOrEnv,
    parseConfiguratorAddress,
    pullSettingFromConfigurator,
)


class ConfigSource(Enum):
    """
    Enum for configs source
    """

    # settings was load from defaults
    default = "default"  # pylint: disable-msg=C0103
    # settings was load from configurator
    configurator = "configurator"  # pylint: disable-msg=C0103
    # settings was load from configuration file
    configFile = "configuration file"  # pylint: disable-msg=C0103


# SETTING TYPE
SETTING_VALUE_TYPE = Union[dict, str, int, float, list, bool]  # pylint: disable-msg=invalid-name


def _updateSettingFromDefaults(defaultValue, settingValue):
    """
    Add missing settings form default

    Args:
        defaultValue: default setting
        settingValue: setting from config file
    """
    settingsDiff = set(defaultValue) - set(settingValue)
    for settingName in settingsDiff:
        settingValue[settingName] = defaultValue[settingName]


def getSchemaFromSetting(setting: Union[dict, BaseSettingDefinition]) -> dict:
    """Get setting validation schema

    Returns:
        object:
    """
    if isinstance(setting, dict):
        return setting["validation_schema"]
    return setting.schema


def getDefaultFromSetting(setting: Union[dict, BaseSettingDefinition]) -> SETTING_VALUE_TYPE:
    """Get setting default value"""
    if isinstance(setting, dict):
        return setting["default_value"]
    return setting.default


@dataclass
class _Config:
    """Config class"""

    # skip or not unknown settings and additional fields
    skipExtras: bool = False


@dataclass
class SettingGetter:
    """Setting getter. Deserialize setting (from file), add default values and validate setting"""

    # setting name
    settingName: str
    # loaded value
    value: SETTING_VALUE_TYPE
    # setting schema
    schema: Union[dict, list]
    # default value
    defaultValue: SETTING_VALUE_TYPE
    # setting source (need for correct deserializing setting from file)
    source: ConfigSource
    # getter setting
    config: _Config = field(default_factory=_Config)

    def getSetting(self) -> SETTING_VALUE_TYPE:
        """Handle loaded value and validate it"""
        setting = self._generateServiceSetting()
        self.validate(setting)
        return setting

    def _getConvertedValue(  # pylint: disable-msg=too-many-branches
        self, schema: dict, value: str
    ) -> Union[str, list, float, int, bool, None]:
        """
        Using data types from the JSON schema to the settings value
        Args:
            schema: data type from JSON schema
            value: setting value

        Returns:
            Value with a given type
        Raises:
            ValueError: Can not convert value to type
            TypeError: Value can not be a list
        """
        valueSchemaType = schema["type"]
        if isinstance(value, str):
            pass  # if setting from the configurator patched with some value from env it may need to be converted
        else:
            return value

        if isinstance(valueSchemaType, list):
            # jsonchema supports several types for value ["array", "string", ...]
            # also nullable field should  be declare as "type": ["array", "null"]
            # here we implemeted only `nullable field` logic
            # https://json-schema.org/understanding-json-schema/reference/type.html
            if len(valueSchemaType) > 2:
                raise NotImplementedError(f"Processing is not implemented for '{valueSchemaType}'")
            if "null" not in valueSchemaType:
                raise NotImplementedError(f"Processing is not implemented for '{valueSchemaType}'")
            if value == "" or value is None:
                newValue = None
                return newValue
            valueSchemaType = valueSchemaType[1 - valueSchemaType.index("null")]
        try:
            if valueSchemaType == "string":
                newValue = str(value)  # type: ignore
            elif valueSchemaType == "boolean":
                if value not in ("0", "1"):
                    raise ValueError(f"Can not convert {value} to boolean: 0 or 1 is expected")
                newValue = bool(int(value))  # type: ignore
            elif valueSchemaType == "integer":
                newValue = int(value)  # type: ignore
            elif valueSchemaType == "number":
                newValue = float(value)  # type: ignore
            elif valueSchemaType == "array":
                if isinstance(value, list):
                    listValues = value[:]
                else:
                    if value[0] != "[" or value[-1] != "]":
                        raise TypeError(f"{value} is not list")
                    listValues = map(lambda x: x.strip(), value[1:-1].split(","))
                    listValues = filter(lambda item: item != "", listValues)  # type: ignore
                newValue = [self._getConvertedValue(schema["items"], item) for item in listValues]  # type: ignore
            elif valueSchemaType == "object":
                newValue = {  # type: ignore
                    key: self._getConvertedValue(schema["properties"][key], item)
                    for key, item in value.items()  # type: ignore
                }
            elif valueSchemaType == "null":
                if value not in ("", None):
                    raise ValueError(f"Can not convert {value} to null")
                newValue = None
            else:
                raise NotImplementedError(f"Processing is not implemented for '{valueSchemaType}'")
        except ValueError as exc:
            raise ValueError(f"Can not convert {value} to {valueSchemaType}") from exc
        return newValue

    def isConfigFromConfigurator(self) -> bool:
        """
        Get config from luna-configurator or (config file + defaults)

        Returns:
            True if source of settings is config file otherwise False
        """
        return self.source == ConfigSource.configurator

    def _getSettingsForOneOfSchemaAndFillFromDefaults(self, configSection: dict, defaultSchemas: dict):
        """
        Adds settings from default if they are absent

        Args:
            configSection - section in config file
            defaultSchemas - JSON schema with default settings
        Raises:
            JsonSchemaException: No suitable oneOf schema for setting
        """
        for schema in defaultSchemas:
            copySettings = deepcopy(configSection)
            defaultSetting = schema["default"]

            try:
                if schema["type"] == "object":
                    copySettings = self._getSettingsForSchema(copySettings, schema)  # type: ignore
                else:
                    copySettings = self._getConvertedValue(schema, copySettings)  # type: ignore
            except (KeyError, ValueError):
                continue
            _updateSettingFromDefaults(defaultSetting, copySettings)
            try:
                fastjsonschema.validate(schema, copySettings)
                return copySettings
            except fastjsonschema.exceptions.JsonSchemaException:
                continue
        raise fastjsonschema.exceptions.JsonSchemaException(f"No suitable oneOf schema for setting {configSection}")

    def validate(self, finalSetting: SETTING_VALUE_TYPE):
        """Validate setting"""
        fastjsonschema.validate(self.schema, finalSetting)

    def _generateServiceSetting(self):
        """
        Generate service settings from input json

        Args:
            inputSettingDict - input json with settings
        """
        setting = self._convertSettingsToSchemaCaseAndFillFromDefault()
        return setting

    def _convertSettingsToSchemaCaseAndFillFromDefault(self):
        """
        Convert path in json to schema case and fill settings from default

        Args:
            inputSettingDict - input json with settings
        """
        schema = self.schema
        value = self.value

        if isinstance(value, dict):
            value = defaultdict(dict) | value
            for envkey, envval in os.environ.items():
                if envkey.upper().startswith(f"VL_SETTINGS.{self.settingName}."):
                    envpath = (envkey.lower() if self.isConfigFromConfigurator() else envkey).split(".")[2:]
                    reduce(getitem, envpath[:-1], value)[envpath[-1].lower()] = envval
            value = dict(value)
        else:
            value = os.getenv(f"VL_SETTINGS.OTHER.{self.settingName}".upper()) or value

        if "oneOf" in schema:
            updatedValue = self._getSettingsForOneOfSchemaAndFillFromDefaults(value, schema["oneOf"])
        else:
            updatedValue = self._getSettingsForSchema(value, schema, path="")

            default = self.defaultValue
            if schema["type"] == "object":
                _updateSettingFromDefaults(default, updatedValue)
            else:
                if updatedValue is None:
                    updatedValue = getDefaultFromSetting(self.defaultValue)
        return updatedValue

    def _getSettingsForSchema(self, settingValue: Union[str, dict], schema: dict, path: str = "") -> SETTING_VALUE_TYPE:
        """
        Trying to convert the value to the expected type from the schema

        Args:
            settingValue: setting value with settings
            path: path by setting dict
            schema: JSON schema
        Raises:
            KeyError: Setting name not found by path

        """

        if schema["type"] != "object":
            try:
                return self._getConvertedValue(schema, settingValue)  # type: ignore
            except (TypeError, ValueError) as exc:
                raise ValueError(f"failed to load value for {path.replace('/', '.')}: {exc}") from exc
        settingValue = deepcopy(settingValue)
        settingNames = list(settingValue)
        for settingName in settingNames:
            upperSettingName = settingName.upper()
            lowerSettingName = settingName.lower()
            if upperSettingName in schema["properties"]:
                toCleanSettingName = lowerSettingName
                newSettingName = upperSettingName
            elif lowerSettingName in schema["properties"]:
                toCleanSettingName = upperSettingName
                newSettingName = lowerSettingName
            else:
                if not self.config.skipExtras and not schema["additionalProperties"]:
                    raise KeyError(f"Setting name {settingName} not found by {path}")
                continue

            sectionPath = path + "/" + newSettingName
            settingValue[newSettingName] = settingValue[settingName]  # type: ignore
            if toCleanSettingName in settingValue:
                del settingValue[toCleanSettingName]  # type: ignore
            settingValue[newSettingName] = self._getSettingsForSchema(  # type: ignore
                settingValue[newSettingName],
                schema["properties"][newSettingName],
                sectionPath,
            )
        return settingValue


def createJsonConfigFromConfigFile(loadedConfig) -> dict:
    """
    Creates JSON (python-dict) config from settings .conf file
    and adds standard values

    Returns:
        prepared and updated from all_settings.py config
    """
    if not isinstance(loadedConfig, configparser.ConfigParser):
        raise RuntimeError("Config was not load from defaults or configuration file")
    settingDict = {}

    for section in loadedConfig.sections():
        if section == "OTHER":
            for settingName in loadedConfig[section]:
                settingName = settingName.upper()
                settingValue = loadedConfig[section][settingName]
                settingDict[settingName] = settingValue
        else:
            # it is required to check digits due to diff from
            # https://github.com/dpath-maintainers/dpath-python/compare/v2.1.0...v2.1.1
            path = [int(value) if value.isdigit() else value for value in section.split(".")]
            dpath.util.new(obj=settingDict, path=path, value=dict(loadedConfig[section]))

    return settingDict


def createJsonConfigFromConfigurator(loadedConfig: Optional[dict]) -> dict:
    """
    Prepares data from the configurator for validation according to the scheme

    Returns:
        prepared config JSON format
    """
    if loadedConfig is None or isinstance(loadedConfig, configparser.ConfigParser):
        raise RuntimeError("Config was not pulled from configurator")
    return loadedConfig["value"]


SETTING_MODEL = TypeVar("SETTING_MODEL")  # pylint:  disable-msg=invalid-name


class TypeModel(Generic[SETTING_MODEL]):
    """
    Type Model class.

    This class is designed for declaration non object settings (string, integer) with pydantic.

    Attributes:
        type: type of settings. Built-in types, pytdantic types, may be use.
        _model: generated pydantic model
        default: default settings value

    >>> from typing import Literal
    >>> TypeModel(Literal[0, 1, 2], 0)(1)
    ... 1
    """

    def __init__(self, settingType: Type[SETTING_MODEL], default: SETTING_MODEL):
        self.settingType = settingType
        self._model = _get_parsing_type(settingType, type_name=None)  # type: ignore
        self.default = default

    def schema(self) -> dict:
        """Get schema for setting, actual tor Literal or pydantic Field"""
        return self._model.schema()

    def __call__(self, value: Optional[Union[SETTING_MODEL, _MISSING_TYPE]] = MISSING) -> SETTING_MODEL:
        """Load setting value and validate it. If valur is missing will be return default"""
        if value is MISSING:
            return self.default
        return self._model(__root__=value).__root__


class SettingLoader:  # pylint: disable-msg=too-few-public-methods
    """
    Setting loader. Load setting by name from configurator or config file. Setting is loaded and updated from default
    values.

    If source is config file  and config file has section `LUNA_CONFIGURATOR.USE_CONFIGURATOR = 1` configuration will be
    load from Luna-Configurator.

    Attributes:
        source (ConfigSource): config source
        configuratorOrigin (Optional[str]): optional configurator address
        configuratorApi (Optional[int]): optional configurator api version
        configFile (Optional[str]): optional config file path
    """

    def __init__(
        self,
        source: ConfigSource = ConfigSource.default,
        configuratorAddress: Optional[str] = None,
        configFile: Optional[Union[Path, str]] = None,
    ):
        self.source = source
        if configuratorAddress is not None:
            configuratorOrigin, configuratorApi = parseConfiguratorAddress(configuratorAddress)
            self.configuratorOrigin: Optional[str] = configuratorOrigin
            self.configuratorApi: Optional[int] = configuratorApi
        else:
            self.configuratorOrigin = None
            self.configuratorApi = None
        self.configFile = str(configFile)

        if self.source == ConfigSource.configFile:
            if self.configFile is None:
                raise RuntimeError
            self.source = (
                ConfigSource.configurator
                if isConfiguratorEnableInConfigFileOrEnv(self.configFile)
                else ConfigSource.configFile
            )
            if self.source == ConfigSource.configurator:
                loadedConfig = getConfigFromFile(self.configFile)
                configuratorSection = dict(loadedConfig["LUNA_CONFIGURATOR"])
                self.configuratorOrigin, self.configuratorApi = getConfiguratorAddressFromFile(configuratorSection)

        elif source == ConfigSource.configurator and configuratorAddress is None:
            raise RuntimeError

    def load(
        self, settingName: str, model: Union[Type[SETTING_MODEL], TypeModel[SETTING_MODEL]], tags: Optional[str] = None
    ) -> SETTING_MODEL:
        """
        Load setting, update its from default value, deserialize to model.

        Args:
            settingName: setting name
            model: pydantic model or TypeModel. The model must have a default value.
            tags: tags for pulling from configurator
        """
        schema = model.schema()  # type: ignore
        if isinstance(model, TypeModel):
            defaultValue = model.default
        else:
            defaultValue = model().dict(by_alias=True)  # type: ignore

        if self.source == ConfigSource.configurator:
            settings = pullSettingFromConfigurator(
                settingName=settingName,
                tags=tags,
                configuratorOrigin=self.configuratorOrigin,  # type: ignore
                asyncPulling=False,
                configuratorApiVersion=self.configuratorApi,
            )
            settings = createJsonConfigFromConfigurator(settings)  # type: ignore
        elif self.source == ConfigSource.configFile:
            _loadedConfig = getConfigFromFile(self.configFile)
            configFileSettings = createJsonConfigFromConfigFile(_loadedConfig)
            settings = configFileSettings.get(settingName, defaultValue)
        else:
            settings = defaultValue  # type: ignore
        getter = SettingGetter(settingName, settings, schema, defaultValue, self.source)  # type: ignore
        # skip extra fields which is not interested for user model
        getter.config.skipExtras = True
        settings = getter.getSetting()  # type: ignore
        if isinstance(model, TypeModel):
            return model(settings)  # type: ignore
        return model(**settings)
